#include "MyXMLParser.h"
#include <StringTokenizer.h>
#include <Util.h>
#include <cctype>
#include <ctime>

using namespace std;

MyXMLParser::MyXMLParser()
{
  groupsUsed.clear();
  if( !flog ) flog.close();
}


void MyXMLParser::setLogFile( const string& logfile )
{
  flog.open( logfile.c_str(), ios::app );
  if( flog )
    {
      flog << "------------------------------------------------------------------------\n";
      flog << "Run starts: " << Util::TimeAndDate2String() << endl;
      flog << "------------------------------------------------------------------------\n";
    }
  else 
    {
      cerr << "Warning: Cannot write to file: " << logfile << endl;
    }
}

// Simple text file specifies the names of groups
// that must be kept.
void MyXMLParser::setHolyNamesFromFile( const string& file )
{
  ifstream fin( file.c_str() );
  if( !fin  ) throw common::Exception( "Cannot open file: " + file, common::Exception::ERROR );
  char buffer[ MAXLINEWIDTH ];
  while( fin.getline( buffer, MAXLINEWIDTH ) )
    {
      string line = buffer;
      if( line.size() < 1 ) continue;
      if( line[0] == '#' ) continue;
      holyNames[ line ] = 1;
    }
}


void  MyXMLParser::setForbiddenPatternsFromFile( const string& file )
{
  ifstream fin( file.c_str() );
  if( !fin  ) throw common::Exception( "Cannot open file: " + file, common::Exception::ERROR );
  char buffer[ MAXLINEWIDTH ];
  while( fin.getline( buffer, MAXLINEWIDTH ) )
    {
      string line = buffer;
      if( line.size() < 1 ) continue;
      if( line[0] == '#' ) continue;
      if( line.find( "TOXIC:" ) != string::npos )
	{
	  for( int i = 6; i < line.size(); ++i )
	    {
	      if( line[i] == ' ' ) continue;
	      toxic[ line[i] ] = true;
	    }
	}
      else
	{
	  forbiddenPatterns.push_back( line );
	}
    }
}


// Removing spaces of the current string
void MyXMLParser::removeSpaces( string& s )
{
  string res;
  for( int k = 0; k < s.size(); ++k )
    {
      if( k == 0 && s[k] == ' ' ) continue;
      if( s[k] == ' ' ) res += '_';
      else res += s[k];
    }
  s = res;
}

// Removing spaces of the current string
void MyXMLParser::removeDots( string& s )
{
  string res;
  for( int k = 0; k < s.size(); ++k )
    {
      if( s[k] == '.' );
      else res += s[k];
    }
  s = res;
}
  
// Reads the xml file and stores the attributes
void MyXMLParser::readFromFile( const string& file )
{
  bool readHeader = true;
  ifstream fin( file.c_str() );
  if( !fin ) { cerr << "Cannot open file: " << file << endl; return; }
  char buffer[ MAXLINEWIDTH ];
  int depth = 0;
 
  while( fin.getline( buffer, MAXLINEWIDTH ) )
    {
      string line = buffer;
      if( line == "" && readHeader ) { header.push_back( "" ); continue; }
      auto_ptr<StringTokenizer> myTok( new StringTokenizer( line, " " ) );
      while( myTok->hasMoreTokens() )
	{
	  string tok = myTok->nextToken();
	  if( readHeader && tok == "<BRANCH" ) 
	    { 
	      readHeader = false;
	    }
	  if( readHeader )
	    {
	      header.push_back( line );
	      break;
	    }
	  else
	    {
	      if( tok == "<BRANCH" )
		{
		  // Add a new layer
		  ++depth;
		  Node* pNode = new Node();
		  pNode->setType( "BRANCH" );
		  pNode->setDepth( depth );

		  string tmpline = "";
		  while( myTok->hasMoreTokens() ) tmpline += myTok->nextToken() + " ";
		  string delim = "";
		  delim += '"';
		  StringTokenizer tmpTok( tmpline, delim );
		  while( tmpTok.hasMoreTokens() )
		    {
		      string key = tmpTok.nextToken();
		      string val = tmpTok.nextToken();
		      strip( key );
		      strip( val );
		      //		      cout << key << " " << val << endl;
		      if( key == "" || val == "" ) continue;
		      pNode->setAttribute( key, val );
		    }
		  
		  // Mark the current groupname as used
		  string groupname = pNode->getAttribute( "groupname" );
		  removeSpaces( groupname );
		  if( groupname.size() > 0 ) 
		    {
		      if( groupsUsed[ groupname ] == 1 )
			{
			  if( flog ) flog << "WARNING: groupname \""<< groupname << "\" occurs more than ones in infile!\n";
			}
		      groupsUsed[ groupname ] = 1;
		      //cout << groupname << " " << groupsUsed[ groupname ] << endl;
		    }

		  myNodes.push_back( pNode );
		}
	      else if( tok == "</BRANCH>" )
		{
		  // Step up one layer
		  --depth;
		  Node* pNode = new Node();
		  pNode->setType( "/BRANCH" );
		  pNode->setDepth( depth );
		  myNodes.push_back( pNode );
		}
	      else if( tok == "<ITEM" )
		{
		  Node* pNode = new Node();
		  pNode->setType( "ITEM" );
		  pNode->setDepth( depth );
		  string tmpline = "";
		  while( myTok->hasMoreTokens() ) tmpline += myTok->nextToken() + " ";
		  StringTokenizer tmpTok( tmpline, "\"" );
		  while( tmpTok.hasMoreTokens() )
		    {
		      string key = tmpTok.nextToken();
		      string val = tmpTok.nextToken();
		      strip( key );
		      strip( val );
		      if( key == "" || val == "" ) continue;
		      pNode->setAttribute( key, val );
		    }
		  myNodes.push_back( pNode );
		} 
	    }
	}
    }
}


// Remove un-wanted characters from the 
// attributes defined in the xml tags.
void MyXMLParser::strip( string& r )
{
  string res = "";
  for( int i = 0; i < r.size(); ++i )
    {
      if( i == 0 && r[0] == ' ' ) continue; 
      if( r[i] == '>' || r[i] == '"' || r[i] == '=' ) {}
      else res += r[i];
    }
  r = res;
}

// Prints XML format to std output
void MyXMLParser::print() 
{
  for( int i = 0; i < header.size(); ++i )
    {
      cout << header[i] << "\n";
    }
  for( int i = 0; i < myNodes.size(); ++i )
    {
      myNodes[i]->print();
    }
  cout << "</ARB_TREE>\n";
}

// Prints to file
// stripItemname will reduce the itemname tag to its identifier
// given be the first field.
void MyXMLParser::print( ofstream& fout, const bool& stripItemname )
{
  for( int i = 0; i < header.size(); ++i )
    {
      fout << header[i] << "\n";
    }
  for( int i = 0; i < myNodes.size(); ++i )
    {
      myNodes[i]->print( fout, stripItemname );
    }
  fout << "</ARB_TREE>\n";
}

// Prints branch lengths and support values
void MyXMLParser::printLandS( ofstream& fout )
{
  for( int i = 0; i < myNodes.size(); ++i )
    {
      if( myNodes[i]->getType() == "BRANCH" ) 
	{
	  fout << myNodes[i]->getAttribute( "length" ) << " " 
	       << myNodes[i]->getAttribute( "bootstrap" ) << " " 
	       << myNodes[i]->getAttribute( "groupname" ) << "\n"; 
	}
    }
}


// Prints branch lengths and support values
void MyXMLParser::printGroups( ofstream& fout )
{
  for( int i = 0; i < myNodes.size(); ++i )
    {
      if( myNodes[i]->getType() == "BRANCH" ) 
	{
	  string group = myNodes[i]->getAttribute( "groupname" );
	  if( group > "" )
	    {
	      fout << myNodes[i]->getAttribute( "groupname" ) << "\n";
	    } 
	}
    }
}


// Returns the string representation of the month as a number
string MyXMLParser::month2number( const string& m )
{
  if( m == "JAN" ) return "01";
  if( m == "FEB" ) return "02";
  if( m == "MAR" ) return "03";
  if( m == "APR" ) return "04";
  if( m == "MAJ" ) return "05";
  if( m == "JUN" ) return "06";
  if( m == "JUL" ) return "07";
  if( m == "AUG" ) return "08";
  if( m == "SEP" ) return "09";
  if( m == "OCT" ) return "10";
  if( m == "NOV" ) return "11";
  if( m == "DEC" ) return "12";
  return "00";
}

// The annotation procedure is coded in this function. All decisions are 
// made based on mC, mL and mS. Parameters specified by the user. 
int MyXMLParser::addGroupNames( const int& mC, const double& mL, const int& mS )
{
  int numberGroups = 0;
  for( double depth = Node::getMaxDepth(); depth > 0; --depth )
    {
      vector<Node*>::iterator i_nod = myNodes.begin();
      //   cout << "Depth = " << depth << endl;
      for( ; i_nod != myNodes.end(); ++i_nod )
	{
	  if( (*i_nod)->getDepth() == depth )
	    {
	      if( (*i_nod)->getType() == "BRANCH" ) 
		{
		  string bl =  (*i_nod)->getAttribute( "length" );
		  string sup = (*i_nod)->getAttribute( "bootstrap" ) ;
		  if( (*i_nod)->getAttribute( "groupname" ).size() > 0 ) continue;
		  if( atof( bl.c_str() ) < mL && mL >= 0 )  continue;
		  if( atof( sup.c_str() ) < mS && mS >= 0 ) continue;
		  vector<Node*>::iterator i_tmp = i_nod;
		  vector<Node*> subset;
		  getItemsBelow( i_tmp, myNodes.end(), subset );
		  if( subset.size() >= mC ) 
		    {
		      ++numberGroups;
		      string val = selectAttribute( subset );
		      //removeSpaces( val );
		      (*i_nod)->setAttribute( "groupname", val );
		    }
		}
	    }
	}
    }
  return numberGroups;
}


string MyXMLParser::getOldestSequence( const vector<string>& f5, 
				       const vector<string>& type,
				       const vector<string>& date,
				       const string& non_pref_type )
{
  string res = "";
  double MINDATE = 99999999;
  for( int i = 0; i < date.size(); ++i )
    {
      string s = f5[i];
      // First check so does not contain any toxic characters
      if( s.size() < 1 ) continue;
      if( !fulfillRequirements( s ) ) continue;
      if( non_pref_type > "" )
	{
	  if( type[i].find( non_pref_type ) != string::npos ) 
	    {
	      continue;
	    }
	}
      StringTokenizer myTok( date[i], "-" );
      string day, month, year;
      if( myTok.hasMoreTokens() )
	{
	  day = myTok.nextToken();
	}
      if( myTok.hasMoreTokens() )
	{
	  month = month2number( myTok.nextToken() );
	}
      if( myTok.hasMoreTokens() )
	{
	  year = myTok.nextToken();
	}
      string date = year + month + day;
      double ndate = atof( date.c_str() ); 
      if( ndate < MINDATE ) 
	{
	  res = s;
	  MINDATE = ndate;
	}
    }
  return res;
}


// Checks all conditions that groupnames must fulfil
bool MyXMLParser::fulfillRequirements( string& s )
{
  // Remove first characters if spaces
  for( int k = 0; k < s.size(); ++k )
    {
      if( s[0] == ' ' ) s = s.substr( 1, s.size() );
      else break;
    }

  // Check the forbidden patterns
  for( int i = 0; i < forbiddenPatterns.size(); ++i )
    {
      if( forbiddenPatterns[i].find( ":BEGIN:NUM:" ) != string::npos )
	{
	  if( isdigit( s[0] ) )
	    {
	      // Modify string so first character is not a number.
	      // s = "x" + s;
	      // removeSpaces( s );
	      return false;
	    }
	}
      if( s.find( forbiddenPatterns[i] ) != string::npos ) 
	{
	  return false;
	}
    }

  // This is a special check to avoid the start of Nan or nan in names
  // as ARB interprets this as Not a Number and transfroms a group
  // name into the largest negative integer. 
  if( s.size() > 3 && ( s.substr( 0, 3 ) == "Nan" || s.substr( 0, 3 ) == "nan" ) )
    {
      s = "x" + s;
    }
  

  // Check the toxic characters
  for( int k = 0; k < s.size(); ++k )
    {
      if( toxic[ s[k] ] ) { return false; }
    }

  // check that name has not been used
  removeSpaces( s );
  removeDots( s );
  if( groupsUsed[ s ] == 1 ) return false;

  // Word seems to be okay
  return true;
}

// This function selects the correct group name according to 
// expert knowledge.
string MyXMLParser::selectAttribute( const vector<Node*>& v  )
{
  string result = "";
  vector<string> id;
  vector<string> type;
  vector<string> ncbi;
  vector<string> date;
  vector<string> field5;
  for( int i = 0; i < v.size(); ++i )
    {
      StringTokenizer myTok( v[i]->getAttribute( "itemname" ) , "," );
      if( myTok.hasMoreTokens() ) id.push_back( myTok.nextToken() );       else id.push_back( "" );
      if( myTok.hasMoreTokens() ) type.push_back( myTok.nextToken() );     else type.push_back( "" );
      if( myTok.hasMoreTokens() ) ncbi.push_back( myTok.nextToken() );     else ncbi.push_back( "" );
      if( myTok.hasMoreTokens() ) date.push_back( myTok.nextToken() );     else date.push_back( "" );
      if( myTok.hasMoreTokens() ) field5.push_back( myTok.nextToken() );   else field5.push_back( "" );
    }
  if( v.size()  > 0 ) 
    {
      map<string, int> myMap;
      
      // Rule 1: Use the taxonomy name of the most common isolate  
      for( int i = 0; i < ncbi.size(); ++i )
	{
	  StringTokenizer myTok( ncbi[i], ";" );
	  string tax = "";
	  // Only used *_isolate for naming
	  while( myTok.hasMoreTokens() ) 
	    {
	      tax = myTok.nextToken();
	    }
	  if( tax.size() < 1 ) continue;

	  // Environmental sample and unclassified should not be used.
	  // These have been hard-coded.
	  if( tax.find( "nvironmental" ) != string::npos ) 
	    {					
	      if( tax.find( "ample" ) != string::npos ) continue;
	    }
	  if( tax.find( "Unclassified" ) != string::npos || 
	      tax.find( "unclassified" ) != string::npos ) continue;
	  
	  // Check and remove groupnames that contain forbidden words
	  if( !fulfillRequirements( tax ) ) continue;
	  myMap[tax]++;
	} 
      
      
      map<string, int>::const_iterator iter;
      int maxCount = 0;
      string groupname = "";
      for( iter = myMap.begin(); iter != myMap.end(); ++iter ) 
	{								
	  if( iter->second > maxCount ) 
	    {
	      maxCount = iter->second;
	      result = iter->first;
	    }
	}
      if( result.size() > 0 )
	{
	  groupsUsed[ result ] = 1;
	  return result;
	}
      
      
      // Rule 2: Use the 5:th field of the oldest clone that fulfill 
      //         requirements of not being "clone"
      result = getOldestSequence( field5, type, date, "clone" );
      if( result.size() > 0 )
	{
	  groupsUsed[ result ] = 1;
	  return result;
	}

      // Rule 3: Use the 5:th field of any record that fulfill 
      //         requirements
      result = getOldestSequence( field5, type, date );
      if( result.size() > 0 )
	{
	  groupsUsed[ result ] = 1;
	  return result;
	}
    }
  
  return "UNNAMEABLE_" + Util::ToString( getTimeInSeconds() );
}

// This function will always provide a unique number unless
// running software on different computers with different 
// system-clocks. However, even if, it is unlikely to produce
// the same number ...
int MyXMLParser::getTimeInSeconds()
{
  static int num = -1;
  if( num < 0 )
    {
      num = time( NULL );
    }
  else 
    {
      return ++num;
    }  
}

// The procedure for removing groupnames. All decisions are 
// made based on mL and mS. Parameters specified by the user.  
int MyXMLParser::removeGroupNames( const double& mL, const int& mS )
{
  int numberGroups = 0;
  vector<Node*>::iterator i_nod = myNodes.begin();
  string fp = "";
  for( ; i_nod != myNodes.end(); ++i_nod )
    {
      bool keepNode = true;
      if( (*i_nod)->getType() == "BRANCH" ) 
	{
	  string groupname = (*i_nod)->getAttribute( "groupname" ); 
	  if( groupname.size() < 1 ) continue;
	  if( holyNames[ groupname ] == 1 ) 
	    {
	      if( flog ) flog << "Attempting to remove: " << groupname << " but declared as holy!\n";
	      continue;
	    }
	  bool cont = false;

	  // Check and remove groupnames that contain forbidden words
	  for( int i = 0; i < forbiddenPatterns.size(); ++i )
	    {
	      if( forbiddenPatterns[i].find( ":BEGIN:NUM:" ) != string::npos )
		{
		  if( isdigit( groupname[0] ) )
		    {
		      cont = true;
		      fp = forbiddenPatterns[i];
		      break;
		    }
		}
	      if( groupname.find( forbiddenPatterns[i] ) != string::npos ) 
		{
		  cont = true;
		  fp = forbiddenPatterns[i];
		  break;
		}
	    }
	  if( cont ) 
	    {
	      if( flog ) flog << "The forbidden pattern: \"" << fp << "\" occurs in: "
			     << groupname << " (groupname removed)." << endl;
	      (*i_nod)->setAttribute( "groupname", "" );
	      ++numberGroups;
	      continue;
	    }
	  
	  // check and remove any groupname that contains a forbidden character
	  char tc;
	  cont = false;
	  for( int i = 0; i < groupname.size(); ++i )
	  {
	    if( toxic[ groupname[i] ] )
	      {
		cont = true;
		tc = groupname[i];
		break;
	      } 
	  }
	  if( cont ) 
	    {
	      if( flog ) flog << "The toxic character:  \"" << tc << "\" occurs in: "
			      << groupname << " (groupname removed)." << endl;
	      (*i_nod)->setAttribute( "groupname", "" );
	      ++numberGroups;
	      continue;
	    }



	  string bl = (*i_nod)->getAttribute( "length" ) ;
	  string sup = (*i_nod)->getAttribute( "bootstrap" ) ;
	  if( atof( bl.c_str() ) < mL   && mL >= 0 ) keepNode = false; 
	  if( atof( sup.c_str() ) >= mS && mS >= 0 ) keepNode = false;
	  if( ( mL < 0 && mS < 0 ) || keepNode ) continue;
	  
	  // Remove group!
	  if( flog ) flog << "Removing group: " << groupname << endl;
	  else       cerr << "Removing group: " << groupname << endl;
	  (*i_nod)->setAttribute( "groupname", "" );
	  ++numberGroups;
	}
    }
  return numberGroups;  
}



// Get all items in a subtree where all nodes are at a deeper level than the current
void MyXMLParser::getItemsBelow( vector<Node*>::iterator it_all, vector<Node*>::iterator it_end, vector<Node*>& subset )
{
  const double depth = (*it_all)->getDepth();
  while( it_all != it_end )
    {
      ++it_all;
      if( (*it_all)->getDepth() <= depth ) break;
      if( (*it_all)->getType() == "BRANCH" && (*it_all)->getAttribute( "groupname" ) > "" )
	{
	  // Groupname already exist. Return an empty list.
	  subset.clear();
	  return;
	}
      if( (*it_all)->getType() == "ITEM" ) subset.push_back( *it_all );
    }
}
