// Copyright (C) 2007 Daniel Dalevi
// Distributed under the GNU GENERAL PUBLIC LICENSE version 2

/*
 * STD includes
 */
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <new>
#include <algorithm>
#include <memory>
#include <cstdlib>
#include <ctime>

using namespace std;

/*
 * Local includes
 */
#include "Parameters.h"
#include "Helper.h"
#include "MyXMLParser.h"

/*
 * Common includes 
 */
#include <Exception.h>
#include <Fasta.h>
#include <StringTokenizer.h>

int main( int argc, char** argv )
{
  Parameters params( true );
  params.LoadParameters( argc, argv );

  if( params.GetBoolParam( "-h" ) == true )
    {
      Helper::displayHelp( argv[0] );
      return 0;
    }
  if( params.GetBoolParam( "-v" ) == true )
    {
      Helper::displayVersion( argv[0] );
      return 0;
    }
  
  auto_ptr<MyXMLParser> myXML( new MyXMLParser() );
  myXML->setLogFile( params.GetStringParam( "-lf" ) );
  myXML->readFromFile( params.GetStringParam( "-f" ) ); 
  if( params.GetStringParam( "-pf" ) > "" )
    {
      try {
	myXML->setForbiddenPatternsFromFile( params.GetStringParam( "-pf" ) );
      }
      catch( common::Exception& o )
	{
	  cerr << o.getText() << endl;
	  return 1;
	}
    }


  if( params.GetBoolParam( "-addGroupNames" ) )
    {
      int mC    = params.GetIntParam( "-mC" );
      double mL = params.GetDoubleParam( "-mL" );
      int mS    = params.GetIntParam( "-mS" );
      int N     = myXML->addGroupNames( mC, mL, mS );
      string outfile = params.GetStringParam( "-o" );
      string outfile_red = "RED_" + outfile;
      ofstream fout( outfile.c_str() ); 
      ofstream fout_red( outfile_red.c_str() ); 
      if( !fout || !fout_red ) { cerr << "Cannot write to output files\n"; return 1; }
      myXML->print( fout, false );
      myXML->print( fout_red, true );
      
      cout << "------------------------------------------------\n" 
	   << "The GrUNT has defined " << N << " new group(s)\n" 
	   << "Each group has: \n"
	   << "(i) At least "<< mC << " number of taxa\n"
	   << "(ii) Branch length > " << mL << endl
	   <<  "(iii) Branch support > " << mS << endl
	   <<  "Input file: " << params.GetStringParam( "-f" ) << endl
	   <<  "Output file: " << outfile << endl
	   <<  "Reduced file: " << outfile_red << endl
	   <<  "Forbidden pattern file: " << params.GetStringParam( "-pf" ) << endl << endl
	   <<  "This is version " << Helper::getVersion() << endl
	   <<  "Latest update was made in " << Helper::getDate() << ".\n"
	   <<  "For help, type, \"GrUNT -h\" \n"
	   <<  "------------------------------------------------\n";
    }
  else if( params.GetBoolParam( "-removeGroupNames" ) )
    {
//       string test = "ADFSFDSDSFDS";
//       if( test.find( "FS" ) != string::npos ) cout << "Found it!\n" << test.find( "FS" )  << endl;
//       return 0;

      // Load holy names
      if( params.GetStringParam( "-hf" ) > "" )
	{
	  try {
	    myXML->setHolyNamesFromFile( params.GetStringParam( "-hf" ) );
	  }
	  catch( common::Exception& o )
	    {
	      cerr << o.getText() << endl;
	      return 1;
	    }
	}
      
      double mL = params.GetDoubleParam( "-mL" );
      int mS    = params.GetIntParam( "-mS" );
      int N     = myXML->removeGroupNames( mL, mS );
      string outfile = params.GetStringParam( "-o" );
      string outfile_red = "RED_" + outfile;
      ofstream fout( outfile.c_str() ); 
      ofstream fout_red( outfile_red.c_str() ); 
      if( !fout || !fout_red ) { cerr << "Cannot write to output files\n"; return 1; }
      myXML->print( fout, false );
      myXML->print( fout_red, true );
      
      cout << "------------------------------------------------\n" 
	   << "The GrUNT has removed " << N << " old group(s)\n" 
	   << "Each group has: \n"
	   << "(i) Branch length < " << mL << endl
	   <<  "(ii) Branch support < " << mS << endl
	   <<  "Input file: " << params.GetStringParam( "-f" ) << endl
	   <<  "Output file: " << outfile << endl
	   <<  "Reduced file: " << outfile_red << endl
	   <<  "Forbidden pattern file: " << params.GetStringParam( "-pf" ) << endl;
      if( params.GetStringParam( "-hf" ) > "" ) cout << "Holy file: " << params.GetStringParam( "-hf" ) << endl;
      cout << endl;
      cout <<  "This is version " << Helper::getVersion() << endl
	   <<  "Latest update was made in " << Helper::getDate() << ".\n"
	   <<  "For help, type, \"GrUNT -h\" \n"
	   <<  "------------------------------------------------\n";
    }
  else if( params.GetBoolParam( "-showLengthAndSupport" ) )
    {
      string outfile = params.GetStringParam( "-o" );
      if( outfile == "out.xml" ) { outfile = "out.txt"; } 
      ofstream fout( outfile.c_str() ); 
      if( !fout ) { cerr << "Cannot write to output files\n"; return 1; }
      myXML->printLandS( fout ); 
    }
  else if( params.GetBoolParam( "-showG" ) )
    {
      string outfile = params.GetStringParam( "-o" );
      if( outfile == "out.xml" ) { outfile = "out.txt"; } 
      ofstream fout( outfile.c_str() ); 
      if( !fout ) { cerr << "Cannot write to output files\n"; return 1; }
      myXML->printGroups( fout ); 
    }
  else 
    {
      Helper::displayShortHelp( argv[0] ); 
    }
  return 0; 
}

